# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import inspect
import logging
import textwrap
from collections import defaultdict
from contextlib import ExitStack, suppress
from functools import partial, wraps
from typing import Any, Callable, ContextManager, TypeVar

from typing_extensions import Protocol

from onetl.exception import SignatureError
from onetl.hooks.hook import CanProcessResult, Hook, HookPriority
from onetl.hooks.hook_collection import HookCollection
from onetl.hooks.hooks_state import HooksState
from onetl.hooks.method_inheritance_stack import MethodInheritanceStack
from onetl.log import NOTICE

Method = TypeVar("Method", bound=Callable[..., Any])

logger = logging.getLogger(__name__)


def _unwrap_method(method: Method) -> Method:
    """Unwrap @classmethod and @staticmethod to get original function"""
    return getattr(method, "__func__", method)


def get_hooks(cls: type, method_name: str) -> HookCollection:
    """Return all hooks registered for a specific method"""

    method = _unwrap_method(cls.__dict__.get(method_name, None))
    return getattr(method, "__hooks__", HookCollection())


def get_hooks_hierarchy(cls: type, method_name: str) -> HookCollection:
    """
    Return all hooks registered for a specific method,
    sorted by priority and class nested level.

    See :ref:`hooks-design` for more details.
    """

    hooks_by_priority: dict[tuple[HookPriority, int], HookCollection] = defaultdict(HookCollection)

    # get all the hooks from the root class to the most nested one
    for nested_level, klass in enumerate(reversed(cls.mro())):
        if method_name not in dir(klass):
            continue

        for hook in get_hooks(klass, method_name).active:
            hooks_by_priority[(hook.priority, nested_level)].add(hook)

    result = HookCollection()
    for priority in sorted(hooks_by_priority):
        result.extend(hooks_by_priority[priority])

    return result


def bind_hook(method: Callable, inp=None):
    """
    Bind a hook to the slot.

    See :ref:`hooks-design` for more details.

    .. versionadded:: 0.7.0

    Examples
    --------

    .. code:: python

        from onetl.hooks import support_hooks, slot, hook, HookPriority


        @support_hooks
        class MyClass:
            @slot
            def method(self, arg):
                pass


        @MyClass.method.bind
        @hook
        def callable(self, arg):
            if arg == "some":
                do_something()


        @MyClass.method.bind
        @hook(priority=HookPriority.FIRST, enabled=True)
        def another_callable(self, arg):
            if arg == "another":
                raise NotAllowed()


        obj = MyClass()
        obj.method(1)  # will call both callable(obj, 1) and another_callable(obj, 1)
    """

    def inner_wrapper(hook):  # noqa: WPS430
        if not isinstance(hook, Hook):
            raise TypeError(
                f"@{method.__qualname__}.bind decorator can be used only on top function marked with @hook",
            )

        method.__hooks__.add(hook)

        logger.debug(
            "|onETL| Registered hook '%s.%s' for '%s' (enabled=%r, priority=%s)",
            hook.__module__,
            hook.__qualname__,  # type: ignore[attr-defined]
            method.__qualname__,
            hook.enabled,
            hook.priority,
        )

        return hook

    if inp is None:
        return inner_wrapper

    return inner_wrapper(inp)


def _prepare_hook_args(
    method: Callable,
    hook: Hook,
    args: tuple,
    kwargs: dict,
) -> inspect.BoundArguments:
    """
    Prepare args for calling hook using original method's args.

    For example, if method has signature:

    .. code:: python

        @support_hooks
        class MyClass:
            @slot
            def method(self, some, named="abc"): ...

    then hook should have a compatible signature, like these ones:

    .. code:: python

        @MyClass.method.bind
        @hook
        def callback(self, some, named): ...

    .. code:: python

        @MyClass.method.bind
        @hook
        def callback(self, some, **kwargs): ...

    .. code:: python

        @MyClass.method.bind
        @hook
        def callback(my_class_instance, *args, **kwargs): ...

    .. note::

        If hook signature contains arg with name ``method_name``, slot method name will be passed into it:

        .. code:: python

            @MyClass.method.bind
            @hook
            def callback(self, method_name, *args, **kwargs):
                assert method_name == "method"

        This can be useful for adding multiple hooks on the same callback function.
    """
    try:
        hook_signature = inspect.signature(hook.callback)
        if "method_name" in hook_signature.parameters:
            return hook_signature.bind(
                *args,
                method_name=method.__name__,
                **kwargs,
            )

        return hook_signature.bind(*args, **kwargs)
    except TypeError as e:
        method_source_line = 0
        with suppress(OSError):
            method_source_line = inspect.getsourcelines(method)[1]

        hook_source_line = 0
        with suppress(OSError):
            hook_source_line = inspect.getsourcelines(hook.callback)[1]

        raise SignatureError(
            textwrap.dedent(
                f"""
                Error while passing method arguments to a hook.

                Method name: '{method.__module__}.{method.__qualname__}'
                Method source: '{inspect.getsourcefile(method)}:{method_source_line}'
                Method signature:
                    {inspect.signature(method)}

                Hook name: '{hook.callback.__module__}.{hook.callback.__qualname__}'
                Hook source: '{inspect.getsourcefile(hook.callback)}:{hook_source_line}'
                Hook signature:
                    {inspect.signature(hook.callback)}
                """,
            ),
        ) from e


def _execute_hook(hook: Hook, args: inspect.BoundArguments):
    """
    Executes the hook using passed arguments, and returns call result.
    """
    try:
        return hook(*args.args, **args.kwargs)
    except Exception:
        logger.exception(
            "|Hooks| Error while executing a hook '%s.%s'",
            hook.__module__,
            hook.__qualname__,  # type: ignore[attr-defined]
        )
        raise


def _handle_context_result(result: Any, context: CanProcessResult, hook: Hook):
    """
    Calls ``context.process_result(result)`` and handles exception, if any.

    See :obj:`onetl.hooks.hook.hook` for more details.
    """
    try:
        return context.process_result(result)
    except Exception:
        logger.exception(
            "|Hooks| Error while passing method call result to the hook '%s.%s'",
            hook.__module__,
            hook.__qualname__,  # type: ignore[attr-defined]
        )
        raise


def register_slot(cls: type, method_name: str):  # noqa: WPS231, WPS213, WPS212
    """
    Internal callback to register ``SomeClass.some_method`` as a slot.

    It is not applied as decorator because decorators have no access to the class but only to a specific method.

    Also ``@classmethod`` is a descriptor, and it can be called only my accessing the class itself,
    which is not possible within a decorator.

    .. versionadded:: 0.7.0

    Examples
    --------

    .. code:: python

        class MyClass:
            def method(self, arg1):
                pass

            @classmethod
            def class_method(cls):
                pass

            @staticmethod
            def static_method(arg1, arg2):
                pass


        MyClass.method = register_slot(MyClass, "method")
        MyClass.class_method = register_slot(MyClass, "class_method")
        MyClass.static_method = register_slot(MyClass, "static_method")
    """

    # <function MyClass.method>
    # <classmethod object>
    # <staticmethod object>
    method_or_descriptor = cls.__dict__[method_name]

    # <function MyClass.method>
    # <function MyClass.class_method>
    # <function MyClass.static_method>
    original_method = _unwrap_method(method_or_descriptor)

    @wraps(original_method)  # noqa: WPS231, WPS213
    def wrapper(*args, **kwargs):  # noqa: WPS231, WPS213
        with MethodInheritanceStack(cls, method_name) as stack_manager, ExitStack() as context_stack:
            if not HooksState.enabled():
                logger.log(NOTICE, "|Hooks| All hooks are disabled")
                return original_method(*args, **kwargs)

            if stack_manager.level > 0:
                logger.log(
                    NOTICE,
                    "|Hooks| Skipping hooks for '%s' because method is called via super()",
                    original_method.__qualname__,
                )
                return original_method(*args, **kwargs)

            hooks = get_hooks_hierarchy(cls, method_name)
            if not hooks:
                logger.log(NOTICE, "|Hooks| No active hooks for '%s'", original_method.__qualname__)
                return original_method(*args, **kwargs)

            original_indent = 4 * stack_manager.level
            indent = original_indent
            context_results: list[tuple[Hook, Any]] = []
            before_results: list[tuple[Hook, Any]] = []

            hooks_count = len(hooks)
            if hooks_count > 1:
                logger.log(
                    NOTICE,
                    "|Hooks| %s%s hooks registered for '%s.%s'",
                    " " * indent,
                    hooks_count,
                    cls.__qualname__,
                    method_name,
                )

            for i, hook in enumerate(hooks):
                logger.log(
                    NOTICE,
                    "|Hooks| %sCalling hook '%s.%s' (%d of %d)",
                    " " * indent,
                    hook.__module__,
                    hook.__qualname__,  # type: ignore[attr-defined]
                    i + 1,
                    hooks_count,
                )

                hook_args = _prepare_hook_args(
                    method=original_method,
                    hook=hook,
                    args=args,
                    kwargs=kwargs,
                )

                hook_call_result = _execute_hook(hook, hook_args)

                if isinstance(hook_call_result, ContextManager):
                    logger.log(
                        NOTICE,
                        "|Hooks| %sThis is a context manager, entering...",
                        " " * indent,
                    )
                    indent += 2
                    context_results.append((hook, hook_call_result))
                    context_stack.enter_context(hook_call_result)
                elif hook_call_result is not None:
                    before_results.append((hook, hook_call_result))
                    logger.log(
                        NOTICE,
                        "|Hooks| %sHook is finished with returning non-None result",
                        " " * indent,
                    )
                else:
                    logger.log(
                        NOTICE,
                        "|Hooks| %sHook is finished with None result",
                        " " * indent,
                    )

            logger.log(
                NOTICE,
                "|Hooks| %sCalling original method '%s.%s.%s'",
                " " * indent,
                cls.__module__,
                cls.__qualname__,
                method_name,
            )

            result = original_method(*args, **kwargs)
            call_result = "None result" if result is None else "non-None result"

            logger.log(
                NOTICE,
                "|Hooks| %sMethod call is finished with %s",
                " " * indent,
                call_result,
            )

            indent = original_indent

            if before_results:
                before_hook, before_result = before_results[-1]

                call_result = "(None)" if result is None else "(*NOT None*)"
                logger.log(
                    NOTICE,
                    "|Hooks| %sMethod call result %s will be replaced with result of calling the hook '%s.%s'",
                    " " * indent,
                    call_result,
                    before_hook.__module__,
                    before_hook.__qualname__,  # type: ignore[attr-defined]
                )
                result = before_result

            for hook, context in context_results:
                if isinstance(context, CanProcessResult):
                    logger.log(
                        NOTICE,
                        "|Hooks| %sPassing result to 'process_result' method of context manager '%s.%s'",
                        " " * indent,
                        hook.__module__,
                        hook.__qualname__,  # type: ignore[attr-defined]
                    )
                    context_result = _handle_context_result(
                        result=result,
                        context=context,
                        hook=hook,
                    )

                    if context_result is not None:
                        call_result = "(None)" if result is None else "(*NOT* None)"  # noqa: WPS220
                        logger.log(  # noqa: WPS220
                            NOTICE,
                            "|Hooks| %sMethod call result %s is modified by hook!",
                            " " * indent,
                            call_result,
                        )
                        result = context_result  # noqa: WPS220
                indent += 2

            return result

    # adding __hooks__ to wrapper is important to detect them with @support_hooks decorator
    wrapper.__hooks__ = method_or_descriptor.__hooks__  # type: ignore[attr-defined]

    wrapper.skip_hooks = wrapper.__hooks__.skip  # type: ignore[attr-defined]
    wrapper.suspend_hooks = wrapper.__hooks__.stop  # type: ignore[attr-defined]
    wrapper.resume_hooks = wrapper.__hooks__.resume  # type: ignore[attr-defined]
    wrapper.bind = partial(bind_hook, wrapper)  # type: ignore[attr-defined]

    # wrap result back to @classmethod and @staticmethod, if was used
    if isinstance(method_or_descriptor, classmethod):
        return classmethod(wrapper)
    elif isinstance(method_or_descriptor, staticmethod):
        return staticmethod(wrapper)

    return wrapper


def _is_method(method: Callable) -> bool:
    """Checks whether class method is callable"""
    method = _unwrap_method(method)
    return callable(method)


def _is_private(method: Callable) -> bool:
    """Checks whether class method is private"""
    method_name = _unwrap_method(method).__name__
    return method_name.startswith("_") and not method_name.endswith("__")


def is_slot(method: Callable) -> bool:
    """Returns True if method is a hook slot"""
    return hasattr(method, "__hooks__")


class Slot(Protocol):
    """Protocol which is implemented by a method after applying :obj:`~slot` decorator.

    .. versionadded:: 0.7.0
    """

    def __call__(self, *args, **kwargs): ...

    @property
    def __hooks__(self) -> HookCollection:
        """Collection of hooks bound to the slot"""

    def skip_hooks(self) -> ContextManager[None]:
        """
        Context manager which temporary stops all the hooks bound to the slot.

        .. note::

            If hooks were stopped by :obj:`~suspend_hooks`, they will not be resumed
            after exiting the context/decorated function.
            You should call :obj:`~resume_hooks` explicitly.

        Examples
        --------

        .. tabs::

            .. code-tab:: py Context manager syntax

                from onetl.hooks.hook import hook, support_hooks, slot


                @support_hooks
                class MyClass:
                    @slot
                    def my_method(self, arg):
                        ...


                @MyClass.my_method.bind
                @hook
                def callback1(self, arg):
                    ...

                obj = MyClass()
                obj.my_method(1)  # will call callback1(obj, 1)

                with MyClass.my_method.skip_hooks():
                    obj.my_method(1)  # will NOT call callback1

                obj.my_method(2)  # will call callback1(obj, 2)

            .. code-tab:: py Decorator syntax

                from onetl.hooks.hook import hook, support_hooks, slot


                @support_hooks
                class MyClass:
                    @slot
                    def my_method(self, arg):
                        ...


                @MyClass.my_method.bind
                @hook
                def callback1(self, arg):
                    ...

                @MyClass.my_method.skip_hooks()
                def method_without_hooks(obj, arg):
                    obj.my_method(arg)


                obj = MyClass()
                obj.my_method(1)  # will call callback1(obj, 1)

                method_without_hooks(obj, 1)  # will NOT call callback1

                obj.my_method(2)  # will call callback1(obj, 2)
        """

    def suspend_hooks(self):
        """
        Stop all the hooks bound to the slot.

        Examples
        --------

        .. code:: python

            from onetl.hooks.hook import hook, support_hooks, slot


            @support_hooks
            class MyClass:
                @slot
                def my_method(self, arg): ...


            @MyClass.my_method.bind
            @hook
            def callback1(self, arg): ...


            obj = MyClass()
            obj.my_method(1)  # will call callback1(obj, 1)

            MyClass.my_method.suspend_hooks()
            obj.my_method(1)  # will NOT call callback1
        """

    def resume_hooks(self):
        """
        Resume all hooks bound to the slot.

        .. note::

            If hook is disabled by :obj:`onetl.hooks.hook.Hook.disable`, it will stay disabled.
            You should call :obj:`onetl.hooks.hook.Hook.enable` explicitly.

        Examples
        --------

        .. code:: python

            from onetl.hooks.hook import hook, support_hooks, slot


            @support_hooks
            class MyClass:
                @slot
                def my_method(self, arg): ...


            @MyClass.my_method.bind
            @hook
            def callback1(self, arg): ...


            obj = MyClass()
            obj.my_method(1)  # will call callback1(obj, 1)

            MyClass.my_method.suspend_hooks()
            obj.my_method(1)  # will NOT call callback1

            MyClass.my_method.resume_hooks()
            obj.my_method(2)  # will call callback1(obj, 2)
        """

    @wraps(bind_hook)
    def bind(self): ...


def slot(method: Method) -> Method:
    """
    Decorator which enables hooks functionality on a specific class method.

    Decorated methods get additional nested methods:

        * :obj:`onetl.hooks.slot.Slot.bind`
        * :obj:`onetl.hooks.slot.Slot.suspend_hooks`
        * :obj:`onetl.hooks.slot.Slot.resume_hooks`
        * :obj:`onetl.hooks.slot.Slot.skip_hooks`

    .. note::

        Supported method types are:

            * Regular methods
            * ``@classmethod``
            * ``@staticmethod``

        It is not allowed to use this decorator over ``_private`` and ``__protected`` methods and ``@property``.
        But is allowed to use on ``__dunder__`` methods, like ``__init__``.

    .. versionadded:: 0.7.0

    Examples
    ---------

    .. code:: python

        from onetl.hooks import support_hooks, slot, hook


        @support_hooks
        class MyClass:
            @slot
            def my_method(self, arg): ...

            @slot  # decorator should be on top of all other decorators
            @classmethod
            def class_method(cls): ...

            @slot  # decorator should be on top of all other decorators
            @staticmethod
            def static_method(arg): ...


        @MyClass.my_method.bind
        @hook
        def callback1(self, arg): ...


        @MyClass.class_method.bind
        @hook
        def callback2(cls): ...


        @MyClass.static_method.bind
        @hook
        def callback3(arg): ...


        obj = MyClass()
        obj.my_method(1)  # will execute callback1(obj, 1)
        MyClass.class_method(2)  # will execute callback2(MyClass, 2)
        MyClass.static_method(3)  # will execute callback3(3)
    """

    if hasattr(method, "__hooks__"):
        raise SyntaxError("Cannot place @slot hook twice on the same method")

    original_method = getattr(method, "__wrapped__", method)

    if not _is_method(original_method):
        raise TypeError(f"@slot decorator could be applied to only to methods of class, got {type(original_method)}")

    if _is_private(original_method):
        raise ValueError(f"@slot decorator could be applied to public methods only, got '{original_method.__name__}'")

    method.__hooks__ = HookCollection()  # type: ignore[attr-defined]
    return method



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import os
import textwrap
import warnings
from typing import Any, Optional

from etl_entities.hwm import HWM
from etl_entities.hwm_store import HWMStoreStackManager

from onetl.hwm import Edge
from onetl.log import log_hwm, log_with_indent
from onetl.strategy.base_strategy import BaseStrategy

log = logging.getLogger(__name__)


class HWMStrategy(BaseStrategy):
    hwm: Optional[HWM] = None

    @property
    def current(self) -> Edge:
        if self.hwm and self.hwm.value is not None:
            return Edge(
                value=self.hwm.value,
                including=False,
            )

        return super().current

    def update_hwm(self, value: Any) -> None:
        if self.hwm and value is not None:
            self.hwm.update(value)

    def enter_hook(self) -> None:
        # if HWM is already set (during previous run),
        # try to fetch its value using qualified_name
        self.fetch_hwm()

    def fetch_hwm(self) -> None:
        class_name = self.__class__.__name__
        if not self.hwm:
            # entering strategy context, HWM will be set later by DBReader.run or FileDownloader.run
            log.debug("|%s| HWM will not be fetched, skipping", class_name)
            return

        hwm_store = HWMStoreStackManager.get_current()
        log.info("|%s| Fetching HWM from %s:", class_name, hwm_store.__class__.__name__)
        log_with_indent(log, "name = %r", self.hwm.name)

        result = hwm_store.get_hwm(self.hwm.name)
        if result is None:
            log.warning(
                "|%s| HWM does not exist in %r. ALL ROWS/FILES WILL BE READ!",
                class_name,
                hwm_store.__class__.__name__,
            )
            return

        log.info("|%s| Fetched HWM:", class_name)
        log_hwm(log, result)

        self.validate_hwm_type(self.hwm, result)
        self.validate_hwm_attributes(self.hwm, result, origin=hwm_store.__class__.__name__)

        self.hwm.set_value(result.value)
        if self.hwm != result:
            log.info("|%s| Final HWM:", class_name)
            log_hwm(log, self.hwm)

    def validate_hwm_type(self, current_hwm: HWM, new_hwm: HWM):
        hwm_type = type(current_hwm)

        if not isinstance(new_hwm, hwm_type):
            message = textwrap.dedent(
                f"""
                Cannot cast HWM of type {type(new_hwm).__name__!r} as {hwm_type.__name__!r}.

                Please:
                * Check that you set correct HWM name, it should be unique.
                * Check that your HWM store contains valid value and type for this HWM name.
                """,
            )
            raise TypeError(message)

    def validate_hwm_attributes(self, current_hwm: HWM, new_hwm: HWM, origin: str):
        attributes = [("entity", True), ("expression", False), ("description", False)]

        for attribute, mandatory in attributes:
            if getattr(current_hwm, attribute) != getattr(new_hwm, attribute):
                # exception raised when inside one strategy >1 processes on the same table but with different entities
                # are executed, example: test_postgres_strategy_incremental_hwm_set_twice
                message = textwrap.dedent(
                    f"""
                    Detected HWM with different `{attribute}` attribute.

                    Current HWM:
                        {current_hwm!r}
                    HWM in {origin}:
                        {new_hwm!r}

                    Please:
                    * Check that you set correct HWM name, it should be unique.
                    * Check that attributes are consistent in both code and HWM Store.
                    """,
                )
                if mandatory:
                    raise ValueError(message)

                warnings.warn(message, UserWarning, stacklevel=2)

    def exit_hook(self, failed: bool = False) -> None:
        if not failed:
            self.save_hwm()

    def save_hwm(self) -> None:
        class_name = self.__class__.__name__

        if not self.hwm:
            log.debug("|%s| HWM value is not set, do not save", class_name)
            return

        hwm_store = HWMStoreStackManager.get_current()

        log.info("|%s| Saving HWM to %r:", class_name, hwm_store.__class__.__name__)
        log_hwm(log, self.hwm)

        location = hwm_store.set_hwm(self.hwm)  # type: ignore
        log.info("|%s| HWM has been saved", class_name)

        if location:
            if isinstance(location, os.PathLike):
                log_with_indent(log, "location = '%s'", os.fspath(location))
            else:
                log_with_indent(log, "location = %r", location)

    @classmethod
    def _log_exclude_fields(cls) -> set[str]:
        return super()._log_exclude_fields() | {"hwm"}



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.dialect_mixins.not_support_columns import (
    NotSupportColumns,
)
from onetl.connection.db_connection.dialect_mixins.not_support_df_schema import (
    NotSupportDFSchema,
)
from onetl.connection.db_connection.dialect_mixins.not_support_hint import (
    NotSupportHint,
)
from onetl.connection.db_connection.dialect_mixins.not_support_where import (
    NotSupportWhere,
)
from onetl.connection.db_connection.dialect_mixins.requires_df_schema import (
    RequiresDFSchema,
)
from onetl.connection.db_connection.dialect_mixins.support_columns_list import (
    SupportColumns,
)
from onetl.connection.db_connection.dialect_mixins.support_hint_str import (
    SupportHintStr,
)
from onetl.connection.db_connection.dialect_mixins.support_hwm_expression_str import (
    SupportHWMExpressionStr,
)
from onetl.connection.db_connection.dialect_mixins.support_name_any import (
    SupportNameAny,
)
from onetl.connection.db_connection.dialect_mixins.support_name_with_schema_only import (
    SupportNameWithSchemaOnly,
)
from onetl.connection.db_connection.dialect_mixins.support_where_str import (
    SupportWhereStr,
)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl.file.file_result import FileResult, FileSet
from onetl.impl import FailedRemoteFile, LocalPath, RemoteFile, RemotePath


class DownloadResult(FileResult):
    """
    Representation of file download result.

    Container for file paths, divided into certain categories:

    * :obj:`~successful`
    * :obj:`~failed`
    * :obj:`~skipped`
    * :obj:`~missing`

    .. versionadded:: 0.3.0

    Examples
    --------

    >>> from onetl.file import FileDownloader
    >>> downloader = FileDownloader(local_path="/local", ...)
    >>> download_result = downloader.run(
    ...     [
    ...         "/remote/file1",
    ...         "/remote/file2",
    ...         "/failed/file",
    ...         "/existing/file",
    ...         "/missing/file",
    ...     ]
    ... )
    >>> download_result
    DownloadResult(
        successful=FileSet([
            LocalPath("/local/file1"),
            LocalPath("/local/file2"),
        ]),
        failed=FileSet([
            FailedLocalFile("/failed/file")
        ]),
        skipped=FileSet([
            RemoteFile("/existing/file")
        ]),
        missing=FileSet([
            RemotePath("/missing/file")
        ]),
    )
    """

    successful: FileSet[LocalPath] = Field(default_factory=FileSet)
    "File paths (local) which were downloaded successfully"

    failed: FileSet[FailedRemoteFile] = Field(default_factory=FileSet)
    "File paths (remote) which were not downloaded because of some failure"

    skipped: FileSet[RemoteFile] = Field(default_factory=FileSet)
    "File paths (remote) which were skipped because of some reason"

    missing: FileSet[RemotePath] = Field(default_factory=FileSet)
    "File paths (remote) which are not present in the remote file system"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

from onetl._util.spark import get_spark_version
from onetl._util.version import Version

if TYPE_CHECKING:
    from py4j.java_gateway import JavaGateway
    from pyspark.sql import SparkSession


def get_java_gateway(spark_session: SparkSession) -> JavaGateway:
    """
    Get py4j Java gateway object
    """
    return spark_session._sc._gateway  # noqa: WPS437 # type: ignore[attr-defined]


def try_import_java_class(spark_session: SparkSession, name: str):
    """
    Try to import Java class by name
    """
    gateway = get_java_gateway(spark_session)
    klass = getattr(gateway.jvm, name)
    gateway.help(klass, display=False)
    return klass


def start_callback_server(spark_session: SparkSession):
    """
    Start Py4J callback server. Important to receive Java events on Python side,
    e.g. in Spark Listener implementations.
    """
    gateway = get_java_gateway(spark_session)
    if get_spark_version(spark_session) >= Version("2.4"):
        from pyspark.java_gateway import ensure_callback_server_started

        ensure_callback_server_started(gateway)
        return

    # PySpark 2.3
    if "_callback_server" not in gateway.__dict__ or gateway._callback_server is None:
        from py4j.java_gateway import JavaObject

        gateway.callback_server_parameters.eager_load = True
        gateway.callback_server_parameters.daemonize = True
        gateway.callback_server_parameters.daemonize_connections = True
        gateway.callback_server_parameters.port = 0
        gateway.start_callback_server(gateway.callback_server_parameters)
        cbport = gateway._callback_server.server_socket.getsockname()[1]
        gateway._callback_server.port = cbport
        # gateway with real port
        gateway._python_proxy_port = gateway._callback_server.port
        # get the GatewayServer object in JVM by ID
        java_gateway = JavaObject("GATEWAY_SERVER", gateway._gateway_client)
        # update the port of CallbackClient with real port
        java_gateway.resetCallbackClient(java_gateway.getCallbackClient().getAddress(), gateway._python_proxy_port)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
# isort: skip_file

from __future__ import annotations

from typing import ClassVar

try:
    from pydantic.v1 import BaseModel as PydanticBaseModel
except (ImportError, AttributeError):
    from pydantic import BaseModel as PydanticBaseModel  # type: ignore[no-redef, assignment]


class BaseModel(PydanticBaseModel):
    _forward_refs_updated: ClassVar[bool] = False

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        extra = "forbid"
        underscore_attrs_are_private = True

    def __init__(self, **kwargs):
        if not self._forward_refs_updated:
            # if pydantic cannot detect type hints (referenced class is not imported yet),
            # it wraps annotation with ForwardRef(...), which should be resolved before creating the instance.
            # so using a small hack to detect all those refs and update them
            # when first object instance is being created
            refs = self._forward_refs()
            self.__class__.update_forward_refs(**refs)
            self.__class__._forward_refs_updated = True  # noqa: WPS437
        super().__init__(**kwargs)

    @classmethod
    def _forward_refs(cls) -> dict[str, type]:
        refs: dict[str, type] = {}
        for item in dir(cls):
            if item.startswith("_") or item.startswith("package"):
                continue

            value = getattr(cls, item)
            if isinstance(value, type):
                refs[item] = value

        return refs
